@@ -10,6 +10,7 @@ const char *ripe_servers[] = {
     "whois.ripe.net",
     "whois.apnic.net",
     "whois.oleane.net",
+    "whois.denic.de",
     "rr.arin.net",		/* does not accept the old syntax */
     "whois.6bone.net",		/* 3.0.0b1 */
     "whois.aunic.net",
@@ -21,7 +22,7 @@ const char *ripe_servers[] = {
     "whois.restena.lu",
     "rr.level3.net",		/* 3.0.0a13 */
     "whois.arnes.si",
-    "rpsl.ripe.net",
+    "www.registry.co.ug",
     NULL
 };
 
@@ -47,14 +48,6 @@ const char *rwhois_servers[] = {
 };
 #endif
 
-const char *gtlds[] = {
-    ".com",
-    ".net",
-    ".org",
-    ".edu",
-    NULL
-};
-
 const char *hide_strings[] = {
     "The Data in Network", "this query",
     "The data in Register", "By submitting",
@@ -64,6 +57,9 @@ const char *hide_strings[] = {
     "Access to ASNIC", "by this policy.",
     "* Copyright (C) 1998 by SGNIC", "* modification.",
     "The Data in Gabia", "you agree to abide",
+    "NeuLevel, Inc., the Registry Operator", "whatsoever, you agree",
+    "NOTICE: Access to .INFO WHOIS", "time. By submitting",
+    "Disclaimer: The Global Name Registry", "for any commercial",
     NULL, NULL
 };
 
