@@ -10,6 +10,13 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <unistd.h>
+#include "config.h"
+#ifdef HAVE_GETOPT_LONG
+#include <getopt.h>
+#endif
+#ifdef HAVE_REGEXEC
+#include <regex.h>
+#endif
 #include <string.h>
 #include <ctype.h>
 #include <sys/types.h>
@@ -19,37 +26,60 @@
 #include <errno.h>
 #include <signal.h>
 
+#undef ENABLE_NLS
+
 /* Application-specific */
-#include "config.h"
 #include "data.h"
 #include "whois.h"
 
 /* Global variables */
-int sockfd, verb = 0, hide_discl = 2;
+int sockfd, verb = 0;
+
+#ifdef ALWAYS_HIDE_DISCL
+int hide_discl = 0;
+#else
+int hide_discl = 2;
+#endif
+
+#ifdef HAVE_GETOPT_LONG
+static struct option longopts[] = {
+    {"help",	no_argument,		NULL, 0  },
+    {"version",	no_argument,		NULL, 1  },
+    {"verbose",	no_argument,		NULL, 'V'},
+    {"server",	required_argument,	NULL, 'h'},
+    {"host",	required_argument,	NULL, 'h'},
+    {"port",	required_argument,	NULL, 'p'},
+    {NULL,	0,			NULL, 0  }
+};
+#else
+extern char *optarg;
+extern int optind;
+#endif
 
 int main(int argc, char *argv[])
 {
-    int ch, nopar = 0, optC = 0;
+    int ch, nopar = 0;
     const char *server = NULL, *port = NULL;
-    char *p, *q, *qstring = NULL, fstring[64] = "\0";
+    char *p, *q, *qstring, fstring[64] = "\0";
 
 #ifdef ENABLE_NLS
-    setlocale(LC_MESSAGES, "");
+    setlocale(LC_ALL, "");
     bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
     textdomain(NLS_CAT_NAME);
 #endif
 
-    while ((ch = getopt(argc, argv, "acCFg:h:Hi:LmMp:rRs:St:T:v:V")) > 0) {
+    while ((ch = GETOPT_LONGISH(argc, argv, "acFg:h:Hi:lLmMp:q:rRs:St:T:v:Vx",
+				longopts, 0)) > 0) {
 	/* RIPE flags */
 	if (strchr(ripeflags, ch)) {
-	    for (p = fstring; *p != '\0'; p++);
+	    for (p = fstring; *p; p++);
 	    sprintf(p--, "-%c ", ch);
 	    continue;
 	}
 	if (strchr(ripeflagsp, ch)) {
-	    for (p = fstring; *p != '\0'; p++);
+	    for (p = fstring; *p; p++);
 	    sprintf(p--, "-%c %s ", ch, optarg);
-	    if (ch == 't' || ch == 'v')
+	    if (ch == 't' || ch == 'v' || ch == 'q')
 		nopar = 1;
 	    continue;
 	}
@@ -57,7 +87,7 @@ int main(int argc, char *argv[])
 	switch (ch) {
 	case 'h':
 	    server = q = malloc(strlen(optarg) + 1);
-	    for (p = optarg; *p != '\0' && *p != ':'; *q++ = tolower(*p++));
+	    for (p = optarg; *p && *p != ':'; *q++ = tolower(*p++));
 	    if (*p == ':')
 		port = p + 1;
 	    *q = '\0';
@@ -68,12 +98,17 @@ int main(int argc, char *argv[])
 	case 'p':
 	    port = optarg;
 	    break;
-	case 'C':
-	    optC = 1;
-	    break;
 	case 'V':
 	    verb = 1;
 	    break;
+	case 1:
+#ifdef VERSION
+	    fprintf(stderr, _("Version %s.\n\nReport bugs to %s.\n"),
+		    VERSION, "<md+whois@linux.it>");
+#else
+	    fprintf(stderr, "%s %s\n", inetutils_package, inetutils_version);
+#endif
+	    exit(0);
 	default:
 	    usage();
 	}
@@ -84,15 +119,16 @@ int main(int argc, char *argv[])
     if (argc == 0 && !nopar)	/* there is no parameter */
 	usage();
 
+    /* On some systems realloc only works on non-NULL buffers */
+    qstring = malloc(1);
+    *qstring = '\0';
+
     /* parse other parameters, if any */
-    if (nopar) {
-	qstring = malloc(1);
-	*qstring = '\0';
-    } else {
+    if (!nopar) {
 	int qslen = 0;
 
 	while (1) {
-	    qslen += strlen(*argv) + 1;
+	    qslen += strlen(*argv) + 1 + 1;
 	    qstring = realloc(qstring, qslen);
 	    strcat(qstring, *argv++);
 	    if (argc == 1)
@@ -102,46 +138,61 @@ int main(int argc, char *argv[])
 	}
     }
 
-    if (!server &&
-#ifndef FIRST_ASK_INTERNIC
-	    optC &&
-#endif
-	    domfind(qstring, gtlds)) {
-	if (verb)
-	    puts(_("Connecting to whois.internic.net."));
-	sockfd = openconn("whois.internic.net", NULL);
-	server = query_crsnic(sockfd, qstring);
-	closeconn(sockfd);
-	if (!server)
-	    exit(0);
-	printf(_("\nFound InterNIC referral to %s.\n\n"), server);
-    }
+    /* -v or -t has been used */
+    if (!server && !*qstring)
+	server = "whois.ripe.net";
 
+#ifdef CONFIG_FILE
     if (!server) {
-	if (!(server = whichwhois(qstring))) {
-	    if (!(server = getenv("WHOIS_SERVER")))
-		server = DEFAULTSERVER;
-	    if (verb)
-		printf(_("Using default server %s.\n"), server);
-	} else if (server[0] == '\001') {
-	    puts(_("This domain has no whois server, but you can access the "
-			"whois database at"));
-	    puts(server + 1);
-	    exit(0);
-	} else if (server[0] == '\002') {
-	    puts(server + 1);
-	    exit(0);
-	} else if (server[0] == '\003') {
-	    puts(_("This domain has no whois server."));
-	    exit(0);
-	} else if (verb)
+	server = match_config_file(qstring);
+	if (verb && server)
 	    printf(_("Using server %s.\n"), server);
     }
+#endif
+
+    if (!server) {
+	char *tmp;
+	tmp = normalize_domain(qstring);
+	server = whichwhois(tmp);
+	free(tmp);
+	switch (server[0]) {
+	    case 0:
+		if (!(server = getenv("WHOIS_SERVER")))
+		    server = DEFAULTSERVER;
+		if (verb)
+		    printf(_("Using default server %s.\n"), server);
+		break;
+	    case 1:
+		puts(_("This TLD has no whois server, but you can access the "
+			    "whois database at"));
+	    case 2:
+		puts(server + 1);
+		exit(0);
+	    case 3:
+		puts(_("This TLD has no whois server."));
+		exit(0);
+	    case 4:
+		if (verb)
+		    puts(_("Connecting to whois.internic.net."));
+		sockfd = openconn("whois.internic.net", NULL);
+		server = query_crsnic(sockfd, qstring);
+		closeconn(sockfd);
+		if (!server)
+		    exit(0);
+		printf(_("\nFound InterNIC referral to %s.\n\n"), server);
+		break;
+	    default:
+		if (verb)
+		    printf(_("Using server %s.\n"), server);
+	}
+    }
+
+    if (getenv("WHOIS_HIDE"))
+	hide_discl = 0;
 
     p = queryformat(server, fstring, qstring);
-    if (verb) {
+    if (verb)
 	printf(_("Query string: \"%s\"\n\n"), p);
-    }
     strcat(p, "\r\n");
 
     signal(SIGTERM, sighandler);
@@ -149,44 +200,119 @@ int main(int argc, char *argv[])
 
     sockfd = openconn(server, port);
     do_query(sockfd, p);
-    closeconn(sockfd);
 
     exit(0);
 }
 
+#ifdef CONFIG_FILE
+const char *match_config_file(const char *s)
+{
+    FILE *fp;
+    char buf[512];
+    static const char delim[] = " \t";
+#ifdef HAVE_REGEXEC
+    regex_t re;
+#endif
+
+    if ((fp = fopen(CONFIG_FILE, "r")) == NULL) {
+	if (errno != ENOENT)
+	    err_sys("Cannot open " CONFIG_FILE);
+	return NULL;
+    }
+
+    while (fgets(buf, sizeof(buf), fp) != NULL) {
+	char *p;
+	const char *pattern, *server;
+#ifdef HAVE_REGEXEC
+	int i;
+#endif
+
+	for (p = buf; *p; p++)
+	    if (*p == '\n')
+		*p = '\0';
+
+	p = buf;
+	while (*p == ' ' || *p == '\t')	/* eat leading blanks */
+	    p++;
+	if (!*p)
+	    continue;		/* skip empty lines */
+	if (*p == '#')
+	    continue;		/* skip comments */
+
+	pattern = strtok(p, delim);
+	server = strtok(NULL, delim);
+	if (!pattern || !server)
+	    err_quit(_("Cannot parse this line: %s"), p);
+	p = strtok(NULL, delim);
+	if (p)
+	    err_quit(_("Cannot parse this line: %s"), p);
+
+#ifdef HAVE_REGEXEC
+	i = regcomp(&re, pattern, REG_EXTENDED|REG_ICASE|REG_NOSUB);
+	if (i != 0) {
+	    char m[1024];
+	    regerror(i, &re, m, sizeof(m));
+	    err_quit("Invalid regular expression '%s': %s", pattern, m);
+	}
+
+	i = regexec(&re, s, 0, NULL, 0);
+	if (i == 0) {
+	    regfree(&re);
+	    return strdup(server);
+	}
+	if (i != REG_NOMATCH) {
+	    char m[1024];
+	    regerror(i, &re, m, sizeof(m));
+	    err_quit("regexec: %s",  m);
+	}
+	regfree(&re);
+#else
+	if (domcmp(s, pattern))
+	    return strdup(server);
+#endif
+    }
+    return NULL;
+}
+#endif
+
 const char *whichwhois(const char *s)
 {
     unsigned long ip;
     unsigned int i;
 
-    /* -v or -t has been used */
-    if (*s == '\0')
-	return "whois.ripe.net";
-
     /* IPv6 address */
     if (strchr(s, ':')) {
-	if (strncasecmp(s, "2001:2", 6) == 0)	/* XXX ugly hack! */
+	if (strncasecmp(s, "2001:2", 6) == 0 ||	/* XXX ugly hack! */
+	    strncasecmp(s, "2001:02", 6) == 0)
 	    return "whois.apnic.net";
-	if (strncasecmp(s, "2001:4", 6) == 0)
+	if (strncasecmp(s, "2001:4", 6) == 0 ||
+	    strncasecmp(s, "2001:04", 6) == 0)
 	    return "whois.arin.net";
-	if (strncasecmp(s, "2001:6", 6) == 0)
+	if (strncasecmp(s, "2001:6", 6) == 0 ||
+	    strncasecmp(s, "2001:06", 6) == 0)
 	    return "whois.ripe.net";
-	//if (strncasecmp(s, "3ffe", 4) == 0)
+	/* if (strncasecmp(s, "3ffe", 4) == 0) */
 	    return "whois.6bone.net";
     }
 
+    /* email address */
+    if (strchr(s, '@'))
+	return "";
+
     /* no dot and no hyphen means it's a NSI NIC handle or ASN (?) */
     if (!strpbrk(s, ".-")) {
 	const char *p;
 
-	for (p = s; *p != '\0'; p++);		/* go to the end of s */
+	for (p = s; *p; p++);			/* go to the end of s */
 	if (strncasecmp(s, "as", 2) == 0 &&	/* it's an AS */
 	    ((s[2] >= '0' && s[2] <= '9') || s[2] == ' '))
 	    return whereas(atoi(s + 2), as_assign);
 	else if (strncasecmp(p - 2, "jp", 2) == 0) /* JP NIC handle */
 	    return "whois.nic.ad.jp";
+	if (*p == '!')	/* NSI NIC handle */
+	    return "whois.networksolutions.com";
 	else /* it's a NSI NIC handle or something we don't know about */
-	    return NULL;
+	    return "";
     }
 
     /* smells like an IP? */
@@ -221,14 +347,14 @@ const char *whichwhois(const char *s)
     if (verb)
 	puts(_("I guess it's a domain but I don't know where to look it"
 	       " up."));
-    return NULL;
+    return "";
 }
 
 const char *whereas(int asn, struct as_del aslist[])
 {
     int i;
 
-    if (asn > 16383)
+    if (asn > 23551)
 	puts(_("Unknown AS number. Please upgrade this program."));
     for (i = 0; aslist[i].serv; i++)
 	if (asn >= aslist[i].first && asn <= aslist[i].last)
@@ -241,7 +367,7 @@ char *queryformat(const char *server, const char *flags, const char *query)
     char *buf;
     int i, isripe = 0;
 
-    /* +10 for CORE; +2 for \r\n; +1 for NULL*/
+    /* +10 for CORE; +2 for \r\n; +1 for NULL */
     buf = malloc(strlen(flags) + strlen(query) + 10 + 2 + 1);
     *buf = '\0';
     for (i = 0; ripe_servers[i]; i++)
@@ -257,7 +383,7 @@ char *queryformat(const char *server, const char *flags, const char *query)
 		isripe = 1;
 		break;
 	    }
-    if (*flags != '\0') {
+    if (*flags) {
 	if (!isripe && strcmp(server, "whois.corenic.net") != 0)
 	    puts(_("Warning: RIPE flags ignored for a traditional server."));
 	else
@@ -272,7 +398,7 @@ char *queryformat(const char *server, const char *flags, const char *query)
     else if (!isripe && strcmp(server, "whois.corenic.net") == 0)
 	sprintf(buf, "--machine %s", query);	/* machine readable output */
     else if (!isripe && strcmp(server, "whois.ncst.ernet.in") == 0 &&
-	    !strchr(query, ' '))
+	     !strchr(query, ' '))
 	sprintf(buf, "domain %s", query);	/* ask for a domain */
     else if (!isripe && strcmp(server, "whois.nic.ad.jp") == 0) {
 	char *lang = getenv("LANG");	/* not a perfect check, but... */
@@ -287,19 +413,18 @@ char *queryformat(const char *server, const char *flags, const char *query)
 
 void do_query(const int sock, const char *query)
 {
-    char buf[200];
+    char buf[2000], *p;
     FILE *fi;
-#ifdef HIDE_DISCL
     int i = 0, hide = hide_discl;
-#endif
 
     fi = fdopen(sock, "r");
     if (write(sock, query, strlen(query)) < 0)
 	err_sys("write");
+/* Using shutdown breaks the buggy RIPE server.
     if (shutdown(sock, 1) < 0)
 	err_sys("shutdown");
-    while (fgets(buf, 200, fi)) {	/* XXX errors? */
-#ifdef HIDE_DISCL
+*/
+    while (fgets(buf, sizeof(buf), fi)) {
 	if (hide == 1) {
 	    if (strncmp(buf, hide_strings[i+1], strlen(hide_strings[i+1]))==0)
 		hide = 2;	/* stop hiding */
@@ -315,7 +440,6 @@ void do_query(const int sock, const char *query)
 	    if (hide == 1)
 		continue;	/* hide the first line */
 	}
-#endif
 #ifdef EXT_6BONE
 	/* % referto: whois -h whois.arin.net -p 43 as 1 */
 	if (strncmp(buf, "% referto:", 10) == 0) {
@@ -329,27 +453,28 @@ void do_query(const int sock, const char *query)
 		strcat(nq, "\r\n");
 		fd = openconn(nh, np);
 		do_query(fd, nq);
-		closeconn(fd);
 		continue;
 	    }
 	}
 #endif
-	fputs(buf, stdout);
+	for (p = buf; *p && *p != '\r' && *p != '\n'; p++);
+	*p = '\0';
+	fprintf(stdout, "%s\n", buf);
     }
     if (ferror(fi))
 	err_sys("fgets");
+    fclose(fi);
 
-#ifdef HIDE_DISCL
     if (hide == 1)
 	err_quit(_("Catastrophic error: disclaimer text has been changed.\n"
 		   "Please upgrade this program.\n"));
-#endif
 }
 
 const char *query_crsnic(const int sock, const char *query)
 {
-    char *temp, buf[100], *ret = NULL;
+    char *temp, buf[2000], *ret = NULL;
     FILE *fi;
+    int state = 0;
 
     temp = malloc(strlen(query) + 1 + 2 + 1);
     *temp = '=';
@@ -359,10 +484,12 @@ const char *query_crsnic(const int sock, const char *query)
     fi = fdopen(sock, "r");
     if (write(sock, temp, strlen(temp)) < 0)
 	err_sys("write");
-    while (fgets(buf, 100, fi)) {
+    while (fgets(buf, sizeof(buf), fi)) {
 	/* If there are multiple matches only the server of the first record
 	   is queried */
-	if (strncmp(buf, "   Whois Server:", 16) == 0 && !ret) {
+	if (state == 0 && strncmp(buf, "   Domain Name:", 15) == 0)
+	    state = 1;
+	if (state == 1 && strncmp(buf, "   Whois Server:", 16) == 0) {
 	    char *p, *q;
 
 	    for (p = buf; *p != ':'; p++);	/* skip until colon */
@@ -370,6 +497,7 @@ const char *query_crsnic(const int sock, const char *query)
 	    ret = malloc(strlen(p) + 1);
 	    for (q = ret; *p != '\n' && *p != '\r'; *q++ = *p++); /*copy data*/
 	    *q = '\0';
+	    state = 2;
 	}
 	fputs(buf, stdout);
     }
@@ -382,8 +510,9 @@ const char *query_crsnic(const int sock, const char *query)
 
 int openconn(const char *server, const char *port)
 {
-    int s, fd;
+    int fd;
 #ifdef HAVE_GETADDRINFO
+    int i;
     struct addrinfo hints, *res, *ressave;
 #else
     struct hostent *hostinfo;
@@ -396,20 +525,19 @@ int openconn(const char *server, const char *port)
     hints.ai_family = AF_UNSPEC;
     hints.ai_socktype = SOCK_STREAM;
 
-    if ((s = getaddrinfo(server, port ? port : "whois", &hints, &res)) != 0)
-	err_quit("getaddrinfo: %s", gai_strerror(s));
-
+    if ((i = getaddrinfo(server, port ? port : "whois", &hints, &res)) != 0)
+	err_quit("getaddrinfo: %s", gai_strerror(i));
     for (ressave = res; res; res = res->ai_next) {
 	if ((fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol))<0)
 	    continue;		/* ignore */
-	if (connect(fd, res->ai_addr, res->ai_addrlen) == 0)
+	if (connect(fd, (struct sockaddr *)res->ai_addr, res->ai_addrlen) == 0)
 	    break;		/* success */
 	close(fd);
     }
+    freeaddrinfo(ressave);
 
     if (!res)
 	err_sys("connect");
-    freeaddrinfo(ressave);
 #else
     if ((hostinfo = gethostbyname(server)) == NULL)
 	err_quit(_("Host %s not found."), server);
@@ -425,10 +553,10 @@ int openconn(const char *server, const char *port)
 	    err_quit(_("%s/tcp: unknown service"), port);
 	saddr.sin_port = servinfo->s_port;
     }
-    if (connect(fd, &saddr, sizeof(saddr)) < 0)
+    if (connect(fd, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
 	err_sys("connect");
 #endif
-    return (fd);
+    return fd;
 }
 
 void closeconn(const int fd)
@@ -447,25 +575,25 @@ int domcmp(const char *dom, const char *tld)
 {
     const char *p, *q;
 
-    for (p = dom; *p != '\0'; p++); p--;	/* move to the last char */
-    for (q = tld; *q != '\0'; q++); q--;
+    for (p = dom; *p; p++); p--;	/* move to the last char */
+    for (q = tld; *q; q++); q--;
     while (p >= dom && q >= tld && tolower(*p) == *q) {	/* compare backwards */
-	if (q == tld)				/* start of the second word? */
+	if (q == tld)			/* start of the second word? */
 	    return 1;
 	p--; q--;
     }
     return 0;
 }
 
-/* check if dom ends with an element of tldlist[] */
-int domfind(const char *dom, const char *tldlist[])
+char *normalize_domain(const char *dom)
 {
-    int i;
+    char *p, *ret;
 
-    for (i = 0; tldlist[i]; i++)
-	if (domcmp(dom, tldlist[i]))
-	    return 1;
-    return 0;
+    ret = strdup(dom);
+    for (p = ret; *p; p++); p--;	/* move to the last char */
+    for (; *p == '.' || p == ret; p--)	/* eat trailing dots */
+	*p = '\0';
+    return ret;
 }
 
 unsigned long myinet_aton(const char *s)
@@ -489,6 +617,8 @@ void usage(void)
 "-h HOST                connect to server HOST\n"
 "-H                     hide legal disclaimers\n"
 "-i ATTR[,ATTR]...      do an inverse lookup for specified ATTRibutes\n"
+"-x                     exact match [RPSL only]\n"
+"-l                     one level less specific lookup [RPSL only]\n"
 "-L                     find all Less specific matches\n"
 "-M                     find all More specific matches\n"
 "-m                     find first level more specific matches\n"
@@ -499,12 +629,16 @@ void usage(void)
 "-S                     tell server to leave out syntactic sugar\n"
 "-s SOURCE[,SOURCE]...  search the database from SOURCE\n"
 "-T TYPE[,TYPE]...      only look for objects of TYPE\n"
-"-t TYPE                requests template for object of TYPE ('all' for a list)\n"
-"-v TYPE                requests verbose template for object of TYPE\n"
-"-V                     explain what is being done\n\n"
-"Version %s. Please report bugs to %s.\n"
-	), VERSION, "<md+whois@linux.it>");
-    exit(1);
+"-t TYPE                request template for object of TYPE ('all' for a list)\n"
+"-v TYPE                request verbose template for object of TYPE\n"
+"-q [version|sources]   query specified server info [RPSL only]\n"
+"-d                     return DNS reverse delegation objects too [RPSL only]\n"
+"-K                     only primary keys are returned [RPSL only]\n"
+"-V    --verbose        explain what is being done\n"
+"      --help           display this help and exit\n"
+"      --version        output version information and exit\n"
+));
+    exit(0);
 }
 
 
